///////////////////////////////////////////////////////////////////////
// WMV plugin for VirtualDub 1.8.1+
// Copyright (C) 2007-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// IVDXAudioSource implementation

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <vfw.h>
#include <crtdbg.h>

#include "vd2\plugin\vdinputdriver.h"
#include "Unknown.h"
#include "WinFile.h"

#include "DataVector.h"

#include "ASFStreamNode.h"
#include "InputFileWMV.h"
#include "AudioSourceWMV.h"


///////////////////////////////////////////////////////////////////////
//
//						AudioSourceWMV
//
///////////////////////////////////////////////////////////////////////

AudioSourceWMV::AudioSourceWMV(InputFileWMV *const pp, ASFStreamNode *const asn)
	: parentPtr (pp)
	, audio     (asn)
	, dwGuess   (0)
	, errmode   (kErrorModeDecodeAnyway)
{
	// Note: parent won't create us unless audio_stream is valid
	parentPtr->AddRef();
}


AudioSourceWMV::~AudioSourceWMV()
{
	parentPtr->Release();
}


int VDXAPIENTRY AudioSourceWMV::AddRef()
{
	return vdxunknown<IVDXStreamSource>::AddRef();
}


int VDXAPIENTRY AudioSourceWMV::Release()
{
	return vdxunknown<IVDXStreamSource>::Release();
}


void *VDXAPIENTRY AudioSourceWMV::AsInterface(uint32 iid)
{
	if (iid == IVDXAudioSource::kIID)
		return static_cast<IVDXAudioSource *>(this);

#ifdef EXPOSE_V3
	if (iid == IVDXStreamSourceV3::kIID)
		return static_cast<IVDXStreamSourceV3 *>(this);
#endif

	return vdxunknown<IVDXStreamSource>::AsInterface(iid);
}


void VDXAPIENTRY AudioSourceWMV::GetStreamSourceInfo(VDXStreamSourceInfo& srcInfo)
{
	srcInfo.mSampleRate.mNumerator         = audio->dwRate;
	srcInfo.mSampleRate.mDenominator       = audio->dwScale;
	srcInfo.mSampleCount                   = audio->dwPerceivedLength;
	srcInfo.mPixelAspectRatio.mNumerator   = 1;
	srcInfo.mPixelAspectRatio.mDenominator = 1;
}


void AudioSourceWMV::UnscrambleAudio(DWORD& dwStart, uint32& lSamples)
{
	// Check for audio scrambling.

	int iSpan = audio->iSpan;

	if (iSpan > 1)
	{
		// Audio is scrambled (sh*t!).
		//
		// ASF1 seems to scramble data differently than described in the
		// ASF2 specs; it's a simple index reversal.  Assuming each
		// scrambling block is composed of sample[a][b], where
		// b=0...(scramble_chunk_size-1), the unscrambling method is
		// simply sample[b][a] (with index dimensions reversed as well,
		// of course).
		//
		// For now, we'll be lazy, and simply decode one sample at a time,
		// reducing the read to a single index transform.  This REALLY
		// should be changed!

		DWORD scramble_base;
		DWORD scramble_offset;
		DWORD scramble_chunk;

		// Break down sample number into (start of scrambled block) and
		// (offset within block).

		scramble_chunk  = audio->iAudioScrambleChunk;
		scramble_offset = dwStart % (iSpan * scramble_chunk);
		scramble_base   = dwStart - scramble_offset;

		// Check to see if this is a partial block and abort descramble if so

		if ((scramble_base + iSpan * scramble_chunk) <= audio->dwActualLength)
		{
			// Descramble offset within block.

			int i = scramble_offset / iSpan;
			int j = scramble_offset % iSpan;

			scramble_offset = i + j * scramble_chunk;

			// Combine to new sample number.

			dwStart = scramble_base + scramble_offset;
			lSamples = 1;
		}
	}
}


bool VDXAPIENTRY AudioSourceWMV::Read(sint64 lStart64, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead)
{
	DWORD   dwStart;
	DWORD	dwActualBytes;
	DWORD   dwActualSamples;

	dwStart = (DWORD)lStart64;

//_RPT4(_CRT_WARN, "AudioSourceWMV::Read %lu, %lu, 0x%p, %lu\n", dwStart, lCount, lpBuffer, cbBuffer);

	if (dwStart >= audio->dwPerceivedLength || lCount == 0)
	{
		if (lBytesRead) *lBytesRead = 0;
		if (lSamplesRead) *lSamplesRead = 0;
		return true;
	}

	// Clamp lCount
	if (lCount > (audio->dwPerceivedLength - dwStart))
	{
		lCount =  audio->dwPerceivedLength - dwStart;
	}

	// Handle repeating and/or scrambled audio
	UnscrambleAudio(dwStart, lCount);

	// Sigh, clamp again...
	if (dwStart >= audio->dwPerceivedLength || lCount == 0)
	{
		if (lBytesRead) *lBytesRead = 0;
		if (lSamplesRead) *lSamplesRead = 0;
		return true;
	}
	if (lCount > (audio->dwPerceivedLength - dwStart))
	{
		lCount =  audio->dwPerceivedLength - dwStart;
	}

	// Buffer size query?
	if (lpBuffer == NULL)
	{
		dwActualBytes   = 0;
		dwActualSamples = 0;

		while (lCount != 0)
		{
			dwActualBytes += audio->sample_list[dwStart].size;
			++dwActualSamples;
			--lCount;
			++dwStart;
			if (dwStart >= audio->dwActualLength) break;
		}

		if (lBytesRead)   *lBytesRead   = dwActualBytes;
		if (lSamplesRead) *lSamplesRead = dwActualSamples;
		return true;
	}

	dwActualBytes   = 0;
	dwActualSamples = 0;

	while (lCount != 0)
	{
		long lSize = audio->sample_list[dwStart].size;
		if (parentPtr->ReadData(lpBuffer, dwStart, 0, lSize, audio) < lSize) break;
		
		lpBuffer = (char *)lpBuffer + lSize;
		dwActualBytes += lSize;
		++dwActualSamples;
		--lCount;
		++dwStart;
		if (dwStart >= audio->dwActualLength) break;
	}

//_RPT2(_CRT_WARN, "AudioSourceWMV::Read returning lBytesRead: %lu, lSamplesRead: %lu\n", dwActualBytes, dwActualSamples);

	if (lBytesRead)   *lBytesRead   = dwActualBytes;
	if (lSamplesRead) *lSamplesRead = dwActualSamples;

	return true;
}


const void *VDXAPIENTRY AudioSourceWMV::GetDirectFormat()
{
	return audio->pFormat;
}


int VDXAPIENTRY AudioSourceWMV::GetDirectFormatLen()
{
	if (audio->pFormat != NULL) return audio->iFormatLen;
	return 0;
}


IVDXStreamSource::ErrorMode VDXAPIENTRY AudioSourceWMV::GetDecodeErrorMode()
{
	return errmode;
}


void VDXAPIENTRY AudioSourceWMV::SetDecodeErrorMode(IVDXStreamSource::ErrorMode mode)
{
	errmode = mode;
}


bool VDXAPIENTRY AudioSourceWMV::IsDecodeErrorModeSupported(IVDXStreamSource::ErrorMode mode)
{
	// TODO
	return true;
}


bool VDXAPIENTRY AudioSourceWMV::IsVBR()
{
	return true;
}


sint64 VDXAPIENTRY AudioSourceWMV::TimeToPositionVBR(sint64 us)
{
	if (audio->dwActualLength != 0)
	{
		const ASFSampleInfo *asi = &audio->sample_list[dwGuess];

		DWORD dwTimeInMs = (DWORD)(us * 0.001 + 0.5);

		if (asi->pres_time > dwTimeInMs)
		{
			// Current guess is too late; find an earlier sample.
			while (dwGuess != 0)
			{
				++asi;
				--dwGuess;
				if (asi->pres_time <= dwTimeInMs) break;
			}
		}
		else
		{
			// Current guess could be too early; check later samples.
			while (dwGuess < (audio->dwActualLength - 1))
			{
				++asi;
				if (asi->pres_time > dwTimeInMs) break;
				++dwGuess;
			}
		}
	}
	return dwGuess;
}


sint64 VDXAPIENTRY AudioSourceWMV::PositionToTimeVBR(sint64 samples)
{
	if (samples >= 0 && samples < (sint64)audio->dwActualLength)
	{
		return (1000i64 * audio->sample_list[samples].pres_time);
	}
	return (sint64)(audio->Duration * 1000000.0 + 0.5);
}


#ifdef EXPOSE_V3

void VDXAPIENTRY AudioSourceWMV::GetStreamSourceInfoV3(VDXStreamSourceInfoV3& srcInfo)
{
	GetStreamSourceInfo(srcInfo.mInfo);
	srcInfo.mFlags = (audio->bHasVSB?) VDXStreamSourceInfoV3::kFlagVariableSizeSamples: 0;
	srcInfo.mfccHandler = 0;
}

#endif


void VDXAPIENTRY AudioSourceWMV::GetAudioSourceInfo(VDXAudioSourceInfo& info)
{
	info.mFlags = 0;
}

